import JetBrainsMonoRegular from "./JetBrainsMono-Regular.woff2";

import { type ExcalidrawFontFaceDescriptor } from "../Fonts";

export const JetBrainsMonoFontFaces: ExcalidrawFontFaceDescriptor[] = [
  {
    uri: JetBrainsMonoRegular,
  },
];