import AdwaitaSansRegular from "./AdwaitaSans-Regular.woff2";

import { type ExcalidrawFontFaceDescriptor } from "../Fonts";

export const AdwaitaSansFontFaces: ExcalidrawFontFaceDescriptor[] = [
  {
    uri: AdwaitaSansRegular,
  },
];